
ELEMENT_DOCUMENT:      str = 'PyutDocument'
TOP_LEVEL_ELEMENT:     str = 'PyutProject'
ELEMENT_GRAPHIC_CLASS: str = 'GraphicClass'
ELEMENT_GRAPHIC_LINK:  str = 'GraphicLink'
ELEMENT_MODEL_CLASS:   str = 'Class'
ELEMENT_MODEL_LINK:    str = 'Link'
ELEMENT_CONTROL_POINT: str = 'ControlPoint'

ELEMENT_MODEL_METHOD: str = 'Method'

ATTR_WIDTH:  str = 'width'
ATTR_HEIGHT: str = 'height'
ATTR_X:      str = 'x'
ATTR_Y:      str = 'y'
ATTR_NAME:   str = 'name'

ATTR_LINK_SOURCE_ANCHOR_X: str = 'srcX'
ATTR_LINK_SOURCE_ANCHOR_Y: str = 'srcY'

ATTR_LINK_DESTINATION_ANCHOR_X: str = 'dstX'
ATTR_LINK_DESTINATION_ANCHOR_Y: str = 'dstY'

ATTR_TYPE: str = 'type'

ATTR_SHOW_METHODS:      str = 'showMethods'
ATTR_SHOW_FIELDS:       str = 'showFields'
ATTR_SHOW_STEREOTYPE:   str = 'showStereotype'

ATTR_DISPLAY_PARAMETERS: str = 'displayParameters'

ELEMENT_MODEL_PARAM: str = 'Param'
ATTR_DEFAULT_VALUE:  str = 'defaultValue'
